@@ -1464,11 +1464,34 @@ IR_FOLD(EQ(SEXT, C_I32))
 IR_FOLD(EQ(SEXT, C_I64))
 IR_FOLD(EQ(SEXT, C_ADDR))
 {
-	if (op2_insn->val.u64 == 0 && ctx->ir_base[op1_insn->op1].type == IR_BOOL) {
+	if (ctx->use_lists && ctx->use_lists[op1_insn->op1].count != 1) {
+		/* pass */
+	} else if (op2_insn->val.u64 == 0 && ctx->ir_base[op1_insn->op1].type == IR_BOOL) {
 		opt = IR_OPT(IR_NOT, IR_BOOL);
 		op1 = op1_insn->op1;
 		op2 = IR_UNUSED;
 		IR_FOLD_RESTART;
+	} else {
+		ir_type type = ctx->ir_base[op1_insn->op1].type;
+
+		if (IR_IS_TYPE_SIGNED(type)) {
+			switch (ir_type_size[type]) {
+				case 1:  val.i64 = op2_insn->val.i8;  break;
+				case 2:  val.i64 = op2_insn->val.i16; break;
+				case 4:  val.i64 = op2_insn->val.i32; break;
+				default: val.u64 = op2_insn->val.u64; break;
+			 }
+	    } else {
+			switch (ir_type_size[type]) {
+				case 1:  val.u64 = op2_insn->val.u8;  break;
+				case 2:  val.u64 = op2_insn->val.u16; break;
+				case 4:  val.u64 = op2_insn->val.u32; break;
+				default: val.u64 = op2_insn->val.u64; break;
+			 }
+		}
+		op1 = op1_insn->op1;
+		op2 = ir_const(ctx, val, type);
+		IR_FOLD_RESTART;
 	}
 	IR_FOLD_NEXT;
 }
@@ -1488,8 +1511,31 @@ IR_FOLD(NE(SEXT, C_I32))
 IR_FOLD(NE(SEXT, C_I64))
 IR_FOLD(NE(SEXT, C_ADDR))
 {
-	if (op2_insn->val.u64 == 0 && ctx->ir_base[op1_insn->op1].type == IR_BOOL) {
+	if (ctx->use_lists && ctx->use_lists[op1_insn->op1].count != 1) {
+		/* pass */
+	} else if (op2_insn->val.u64 == 0 && ctx->ir_base[op1_insn->op1].type == IR_BOOL) {
 		IR_FOLD_COPY(op1_insn->op1);
+	} else {
+		ir_type type = ctx->ir_base[op1_insn->op1].type;
+
+		if (IR_IS_TYPE_SIGNED(type)) {
+			switch (ir_type_size[type]) {
+				case 1:  val.i64 = op2_insn->val.i8;  break;
+				case 2:  val.i64 = op2_insn->val.i16; break;
+				case 4:  val.i64 = op2_insn->val.i32; break;
+				default: val.u64 = op2_insn->val.u64; break;
+			 }
+	    } else {
+			switch (ir_type_size[type]) {
+				case 1:  val.u64 = op2_insn->val.u8;  break;
+				case 2:  val.u64 = op2_insn->val.u16; break;
+				case 4:  val.u64 = op2_insn->val.u32; break;
+				default: val.u64 = op2_insn->val.u64; break;
+			 }
+		}
+		op1 = op1_insn->op1;
+		op2 = ir_const(ctx, val, type);
+		IR_FOLD_RESTART;
 	}
 	IR_FOLD_NEXT;
 }
@@ -1586,6 +1632,24 @@ IR_FOLD(SUB_OV(_, C_ADDR))
 	IR_FOLD_NEXT;
 }
 
+/* This rule is useful for ADD(0, SYM) => SYM */
+IR_FOLD(ADD(C_U8, _))
+IR_FOLD(ADD(C_U16, _))
+IR_FOLD(ADD(C_U32, _))
+IR_FOLD(ADD(C_U64, _))
+IR_FOLD(ADD(C_I8, _))
+IR_FOLD(ADD(C_I16, _))
+IR_FOLD(ADD(C_I32, _))
+IR_FOLD(ADD(C_I64, _))
+IR_FOLD(ADD(C_ADDR, _))
+{
+	if (op1_insn->val.u64 == 0) {
+		/* 0 + a => a */
+		IR_FOLD_COPY(op2);
+	}
+	IR_FOLD_NEXT;
+}
+
 IR_FOLD(SUB(C_I8, _))
 IR_FOLD(SUB(C_I16, _))
 IR_FOLD(SUB(C_I32, _))
@@ -2404,6 +2468,17 @@ IR_FOLD(SEXT(AND))
 	IR_FOLD_NEXT;
 }
 
+IR_FOLD(SEXT(SHR))
+{
+	if (IR_IS_CONST_REF(op1_insn->op2)
+	 && !IR_IS_SYM_CONST(ctx->ir_base[op1_insn->op2].op)
+	 && ctx->ir_base[op1_insn->op2].val.u64 != 0) {
+		opt = IR_OPT(IR_ZEXT, IR_OPT_TYPE(opt));
+		IR_FOLD_RESTART;
+	}
+	IR_FOLD_NEXT;
+}
+
 IR_FOLD(TRUNC(AND))
 {
 	if (IR_IS_CONST_REF(op1_insn->op2)) {
@@ -2430,6 +2505,44 @@ IR_FOLD(TRUNC(AND))
 	IR_FOLD_NEXT;
 }
 
+IR_FOLD(AND(ZEXT, C_I16))
+IR_FOLD(AND(ZEXT, C_U16))
+IR_FOLD(AND(ZEXT, C_I32))
+IR_FOLD(AND(ZEXT, C_U32))
+IR_FOLD(AND(ZEXT, C_I64))
+IR_FOLD(AND(ZEXT, C_U64))
+IR_FOLD(AND(ZEXT, C_ADDR))
+{
+	ir_type src_size = ir_type_size[ctx->ir_base[op1_insn->op1].type];
+
+	if ((src_size == 1 && op2_insn->val.u64 == 0xff)
+	 || (src_size == 2 && op2_insn->val.u64 == 0xffff)
+	 || (src_size == 4 && op2_insn->val.u64 == 0xffffffff)) {
+		IR_FOLD_COPY(op1);
+	}
+	IR_FOLD_NEXT;
+}
+
+IR_FOLD(AND(SEXT, C_I16))
+IR_FOLD(AND(SEXT, C_U16))
+IR_FOLD(AND(SEXT, C_I32))
+IR_FOLD(AND(SEXT, C_U32))
+IR_FOLD(AND(SEXT, C_I64))
+IR_FOLD(AND(SEXT, C_U64))
+IR_FOLD(AND(SEXT, C_ADDR))
+{
+	ir_type src_size = ir_type_size[ctx->ir_base[op1_insn->op1].type];
+
+	if ((src_size == 1 && op2_insn->val.u64 == 0xff)
+	 || (src_size == 2 && op2_insn->val.u64 == 0xffff)
+	 || (src_size == 4 && op2_insn->val.u64 == 0xffffffff)) {
+		opt = IR_OPT(IR_ZEXT, IR_OPT_TYPE(opt));
+		op1 = op1_insn->op1;
+		op2 = IR_UNUSED;
+		IR_FOLD_RESTART;
+	}
+	IR_FOLD_NEXT;
+}
 IR_FOLD(AND(SHR, C_I8))
 IR_FOLD(AND(SHR, C_U8))
 {
@@ -3013,6 +3126,7 @@ IR_FOLD(UGT(_, _))
 	} else if (op1 < op2) {  /* move lower ref to op2 */
 		SWAP_REFS(op1, op2);
 		opt ^= 3; /* [U]LT <-> [U]GT, [U]LE <-> [U]GE */
+		IR_FOLD_RESTART;
 	}
 	IR_FOLD_NEXT;
 }
